import { d } from './dice';
import { individualTreasureData } from './treasureData';

class TreasureService {
  private individualRewardsTables: RandomIndividualTreasureTable[] = [];
  private tableByName: {[tableName: string]: RandomIndividualTreasureTable} = {};
  private tablesByTags: {[tagName: string]: RandomIndividualTreasureTable[]} = {};
  private tags: Set<string> = new Set();

  constructor() {
    console.time('TreasureServiceStartup');
    for (let i in individualTreasureData) {
      let table = individualTreasureData[i];
      let randomTable: RandomIndividualTreasureTable = new RandomIndividualTreasureTable(table);
      this.individualRewardsTables.push(randomTable);
      this.tableByName[table.name] = randomTable;
      if (table.tags) {
        for (let j in table.tags) {
          let tag: string = table.tags[j];
          if (!this.tablesByTags[tag]) {
            this.tablesByTags[tag] = [];
          }
          this.tags.add(tag);
          this.tablesByTags[tag].push(randomTable);
        }
      }
    }
    console.timeEnd('TreasureServiceStartup');
  }

  public getTablesWithTag(tag: string) {
    return this.tablesByTags[tag];
  }

  public getTable(name: string) {
    return this.tableByName[name];
  }

  public getTags() {
    return this.tags;
  }

  public searchForTable(search: string) {
    let answer = this.individualRewardsTables.filter((table: RandomIndividualTreasureTable) => {
      return table.name.toLowerCase().indexOf(search.toLowerCase()) !== -1 || table.getTags().join(' ').toLowerCase().indexOf(search.toLowerCase()) !== -1;
    }).map((table: RandomIndividualTreasureTable) => {
      let score = search.length / table.name.length;
      return {
        table: table,
        score: score
      };
    }).sort((a: {table: RandomIndividualTreasureTable, score: number}, b: {table: RandomIndividualTreasureTable, score: number}) => {
      return a.score - b.score;
    }).map((tableScore: {table: RandomIndividualTreasureTable, score: number}) => {
      return tableScore.table;
    });

    return answer;
  }
}

class RandomIndividualTreasureTable {
  public name: string;
  public dice: number[];
  private indexedResults: CoinRewards[] = [];
  private config: IndividualTreasureTable;

  constructor(config: IndividualTreasureTable) {
    this.name = config.name;
    this.dice = config.dice;
    this.constructIndex(config.possibleCoinRewards);
    this.config = config;
  }

  public roll(input?: number) {
    let testRoll: number = this.getRoll();
    if (!this.isValidInput(testRoll)) {
      console.log(testRoll);
      debugger;
    }
    if (!input || !this.isValidInput(input)) {
      input = this.getRoll();
    }
    return this.indexedResults[input - 1];
  }

  public getPossibleRewards() {
    return this.config.possibleCoinRewards;
  }

  public getTags() {
    return this.config.tags || [];
  }

  public getNotes() {
    return this.config.notes;
  }

  private getRoll(): number {
    let answer: number = 0;
    for (let i in this.dice) {
      answer = answer + d(this.dice[i]);
    }
    return answer;
  }

  private isValidInput(input: number) {
    return typeof this.indexedResults[input - 1] !== 'undefined';
  }

  private constructIndex(coinRewards: CoinRewards[]) {
    for (let resultIndex in coinRewards) {
      let keys: number[] = coinRewards[resultIndex].keys;
      for (let r = Math.min(...keys); r <= Math.max(...keys); r++) {
        this.indexedResults.push(coinRewards[resultIndex]);
      }
    }
  }
}

type ChallengeRating = '0-4' | '5-10' | '11-16' | '17+';

interface CoinResult {
  type: 'CP' | 'SP' | 'EP' | 'GP' | 'PP';
  count: number;
}

interface CoinCalc {
  type: 'CP' | 'SP' | 'EP' | 'GP' | 'PP';
  rollCount: number;
  diceType: number;
  multiplier: number;
  avg: number;
}

interface CoinRewards {
  keys: number[];
  coinCalc: CoinCalc[];
}

interface GemArtCalc {
  tableName: string;
  rollCount: number;
  diceType: number;
  avg: number;
}

interface GemArtResult {
  tableName: string;
  amount: number;
}

interface MagicItemCalc {
  tableName: string;
  rollCount: number;
  diceType: number;
}

interface MagicItemResult {
  tableName: string;
  numberOfRoles: number;
}

interface OtherRewards {
  keys: number[];
  gemArt: GemArtCalc;
  magicItem: MagicItemCalc[];
}

interface IndividualTreasureTable {
  name: string;
  notes?: string;
  tags?: string[];
  dice: number[];
  possibleCoinRewards: CoinRewards[];
}

interface HoardTresureTable {
  name: string;
  notes?: string;
  tags?: string[];
  dice: number[];
  coinRewards: CoinRewards;
  possibleRewards: OtherRewards[]
}

const treasureService: TreasureService = new TreasureService();

export {
  HoardTresureTable,
  IndividualTreasureTable,
  RandomIndividualTreasureTable,
  treasureService,
}
